import math
import os
import time
from collections import defaultdict
from pathlib import Path
from typing import Dict, List, Literal, Optional, TypeVar, Union

import numpy as np
import srsly
import torch
from colbert import Indexer, IndexUpdater, Searcher, Trainer
from colbert.infra import ColBERTConfig, Run, RunConfig
from colbert.modeling.checkpoint import Checkpoint

from ragatouille.models.base import LateInteractionModel


class ColBERT(LateInteractionModel):
    def __init__(
        self,
        pretrained_model_name_or_path: Union[str, Path],
        n_gpu: int = -1,
        index_name: Optional[str] = None,
        verbose: int = 1,
        load_from_index: bool = False,
        training_mode: bool = False,
        index_root: Optional[str] = None,
        **kwargs,
    ):
        self.verbose = verbose
        self.collection = None
        self.pid_docid_map = None
        self.docid_metadata_map = None
        self.in_memory_docs = []
        if n_gpu == -1:
            n_gpu = 1 if torch.cuda.device_count() == 0 else torch.cuda.device_count()

        self.loaded_from_index = load_from_index

        if load_from_index:
            self.index_path = str(pretrained_model_name_or_path)
            ckpt_config = ColBERTConfig.load_from_index(
                str(pretrained_model_name_or_path)
            )
            self.config = ckpt_config
            self.run_config = RunConfig(
                nranks=n_gpu, experiment=self.config.experiment, root=self.config.root
            )
            split_root = str(pretrained_model_name_or_path).split("/")[:-1]
            self.config.root = "/".join(split_root)
            self.checkpoint = self.config.checkpoint
            self.index_name = self.config.index_name
            self.collection = self._get_collection_from_file(
                str(pretrained_model_name_or_path / "collection.json")
            )
            try:
                self.pid_docid_map = self._get_collection_from_file(
                    str(pretrained_model_name_or_path / "pid_docid_map.json")
                )
                # convert all keys to int when loading from file because saving converts to str
                self.pid_docid_map = {
                    int(key): value for key, value in self.pid_docid_map.items()
                }
                self.docid_pid_map = defaultdict(list)
                for pid, docid in self.pid_docid_map.items():
                    self.docid_pid_map[docid].append(pid)
                if os.path.exists(
                    str(pretrained_model_name_or_path / "docid_metadata_map.json")
                ):
                    self.docid_metadata_map = self._get_collection_from_file(
                        str(pretrained_model_name_or_path / "docid_metadata_map.json")
                    )
            except Exception:
                print(
                    "WARNING: Could not load pid_docid_map or docid_metadata_map from index!",
                    "This is likely because you are loading an old index.",
                )
                self.pid_docid_map = defaultdict(lambda: None)
                self.docid_metadata_map = defaultdict(lambda: None)
            # TODO: Modify root assignment when loading from HF

        else:
            self.index_root = index_root if index_root is not None else ".ragatouille/"
            ckpt_config = ColBERTConfig.load_from_checkpoint(
                str(pretrained_model_name_or_path)
            )
            self.run_config = RunConfig(
                nranks=n_gpu, experiment="colbert", root=self.index_root
            )
            local_config = ColBERTConfig(**kwargs)
            self.config = ColBERTConfig.from_existing(
                ckpt_config,
                local_config,
            )
            self.checkpoint = pretrained_model_name_or_path
            self.index_name = index_name
            self.config.experiment = "colbert"
            self.config.root = ".ragatouille/"

        if not training_mode:
            self.inference_ckpt = Checkpoint(
                self.checkpoint, colbert_config=self.config
            )

        self.run_context = Run().context(self.run_config)
        self.run_context.__enter__()  # Manually enter the context
        self.searcher = None

    def _get_collection_from_file(self, collection_path: str):
        return srsly.read_json(collection_path)

    def _write_collection_to_file(self, collection, collection_path: str):
        srsly.write_json(collection_path, collection)

    def add_to_index(
        self,
        new_documents: List[str],
        new_pid_docid_map: Dict[int, str],
        new_docid_metadata_map: Optional[List[dict]] = None,
        index_name: Optional[str] = None,
    ):
        self.index_name = index_name if index_name is not None else self.index_name
        if self.index_name is None:
            print(
                "Cannot add to index without an index_name! Please provide one.",
                "Returning empty results.",
            )
            return None

        print(
            "WARNING: add_to_index support is currently experimental!",
            "add_to_index support will be more thorough in future versions",
        )

        if self.loaded_from_index:
            index_root = self.config.root
        else:
            index_root = str(
                Path(self.config.root) / Path(self.config.experiment) / "indexes"
            )
            if not self.collection:
                self.collection = self._get_collection_from_file(
                    str(
                        Path(self.config.root)
                        / Path(self.config.experiment)
                        / "indexes"
                        / self.index_name
                        / "collection.json"
                    )
                )

        searcher = Searcher(
            checkpoint=self.checkpoint,
            config=None,
            collection=self.collection,
            index=self.index_name,
            index_root=index_root,
            verbose=self.verbose,
        )

        current_len = len(searcher.collection)
        new_doc_len = len(new_documents)
        new_documents_with_ids = [
            {"content": doc, "document_id": new_pid_docid_map[pid]}
            for pid, doc in enumerate(new_documents)
            if new_pid_docid_map[pid] not in self.pid_docid_map
        ]

        if new_docid_metadata_map is not None:
            self.docid_metadata_map = self.docid_metadata_map or {}
            self.docid_metadata_map.update(new_docid_metadata_map)

        if current_len + new_doc_len < 5000 or new_doc_len > current_len * 0.05:
            self.index(
                [doc["content"] for doc in new_documents_with_ids],
                {
                    pid: doc["document_id"]
                    for pid, doc in enumerate(new_documents_with_ids)
                },
                docid_metadata_map=self.docid_metadata_map,
                index_name=self.index_name,
                max_document_length=self.config.doc_maxlen,
                overwrite="force_silent_overwrite",
            )
        else:
            updater = IndexUpdater(
                config=self.config, searcher=searcher, checkpoint=self.checkpoint
            )
            updater.add([doc["content"] for doc in new_documents_with_ids])
            updater.persist_to_disk()

        self.pid_docid_map.update(
            {pid: doc["document_id"] for pid, doc in enumerate(new_documents_with_ids)}
        )
        self.docid_pid_map = defaultdict(list)
        for pid, docid in self.pid_docid_map.items():
            self.docid_pid_map[docid].append(pid)

        self._write_collection_to_file(
            self.pid_docid_map, self.index_path + "/pid_docid_map.json"
        )
        if self.docid_metadata_map is not None:
            self._write_collection_to_file(
                self.docid_metadata_map, self.index_path + "/docid_metadata_map.json"
            )

        print(
            f"Successfully updated index with {len(new_documents_with_ids)} new documents!\n",
            f"New index size: {current_len + len(new_documents_with_ids)}",
        )

        self.index_path = str(
            Path(self.run_config.root)
            / Path(self.run_config.experiment)
            / "indexes"
            / self.index_name
        )

        return self.index_path

    def delete_from_index(
        self,
        document_ids: Union[TypeVar("T"), List[TypeVar("T")]],
        index_name: Optional[str] = None,
    ):
        self.index_name = index_name if index_name is not None else self.index_name
        if self.index_name is None:
            print(
                "Cannot delete from index without an index_name! Please provide one.",
                "Returning empty results.",
            )
            return None

        print(
            "WARNING: delete_from_index support is currently experimental!",
            "delete_from_index support will be more thorough in future versions",
        )

        # Initialize the searcher and updater
        searcher = Searcher(
            checkpoint=self.checkpoint,
            config=None,
            collection=self.collection,
            index=self.index_name,
            verbose=self.verbose,
        )
        updater = IndexUpdater(
            config=self.config, searcher=searcher, checkpoint=self.checkpoint
        )

        pids_to_remove = []
        for pid, docid in self.pid_docid_map.items():
            if docid in document_ids:
                pids_to_remove.append(pid)

        updater.remove(pids_to_remove)
        updater.persist_to_disk()

        self.collection = [
            doc for pid, doc in enumerate(self.collection) if pid not in pids_to_remove
        ]
        self.pid_docid_map = {
            pid: docid
            for pid, docid in self.pid_docid_map.items()
            if pid not in pids_to_remove
        }
        self.docid_pid_map = defaultdict(list)
        for pid, docid in self.pid_docid_map.items():
            self.docid_pid_map[docid].append(pid)

        if self.docid_metadata_map is not None:
            self.docid_metadata_map = {
                docid: metadata
                for docid, metadata in self.docid_metadata_map.items()
                if docid not in document_ids
            }

        self._write_collection_to_file(
            self.collection, self.index_path + "/collection.json"
        )
        self._write_collection_to_file(
            self.pid_docid_map, self.index_path + "/pid_docid_map.json"
        )
        if self.docid_metadata_map is not None:
            self._write_collection_to_file(
                self.docid_metadata_map, self.index_path + "/docid_metadata_map.json"
            )

        print(f"Successfully deleted documents with these IDs: {document_ids}")

    def index(
        self,
        collection: List[str],
        pid_docid_map: Dict[int, str],
        docid_metadata_map: Optional[dict] = None,
        index_name: Optional["str"] = None,
        max_document_length: int = 256,
        overwrite: Union[bool, str] = "reuse",
    ):
        if torch.cuda.is_available():
            import faiss

            if not hasattr(faiss, "StandardGpuResources"):
                print(
                    "________________________________________________________________________________\n"
                    "WARNING! You have a GPU available, but only `faiss-cpu` is currently installed.\n",
                    "This means that indexing will be slow. To make use of your GPU.\n"
                    "Please install `faiss-gpu` by running:\n"
                    "pip uninstall --y faiss-cpu & pip install faiss-gpu\n",
                    "________________________________________________________________________________",
                )
                print("Will continue with CPU indexing in 5 seconds...")
                time.sleep(5)
        self.config.doc_maxlen = max_document_length
        if index_name is not None:
            if self.index_name is not None:
                print(
                    f"New index_name received!",
                    f"Updating current index_name ({self.index_name}) to {index_name}",
                )
            self.index_name = index_name
        else:
            if self.index_name is None:
                print(
                    f"No index_name received!",
                    f"Using default index_name ({self.checkpoint}_new_index)",
                )
            self.index_name = self.checkpoint + "new_index"

        self.collection = collection

        nbits = 2
        if len(self.collection) < 5000:
            nbits = 8
        elif len(self.collection) < 10000:
            nbits = 4
        self.config = ColBERTConfig.from_existing(
            self.config, ColBERTConfig(nbits=nbits)
        )

        # Instruct colbert-ai to disable forking if nranks == 1
        self.config.avoid_fork_if_possible = True
        self.indexer = Indexer(
            checkpoint=self.checkpoint,
            config=self.config,
            verbose=self.verbose,
        )
        self.indexer.configure(avoid_fork_if_possible=True)
        self.indexer.index(
            name=self.index_name, collection=self.collection, overwrite=overwrite
        )

        self.index_path = str(
            Path(self.run_config.root)
            / Path(self.run_config.experiment)
            / "indexes"
            / self.index_name
        )
        self.config.root = str(
            Path(self.run_config.root) / Path(self.run_config.experiment) / "indexes"
        )
        self._write_collection_to_file(
            self.collection, self.index_path + "/collection.json"
        )

        self.pid_docid_map = pid_docid_map
        self._write_collection_to_file(
            self.pid_docid_map, self.index_path + "/pid_docid_map.json"
        )

        # inverted mapping for returning full docs
        self.docid_pid_map = defaultdict(list)
        for pid, docid in self.pid_docid_map.items():
            self.docid_pid_map[docid].append(pid)

        if docid_metadata_map is not None:
            self._write_collection_to_file(
                docid_metadata_map, self.index_path + "/docid_metadata_map.json"
            )
            self.docid_metadata_map = docid_metadata_map

        print("Done indexing!")

        return self.index_path

    def _load_searcher(
        self,
        index_name: Optional[str],
        force_fast: bool = False,
    ):
        if index_name is not None:
            if self.index_name is not None:
                print(
                    f"New index_name received!",
                    f"Updating current index_name ({self.index_name}) to {index_name}",
                )
            self.index_name = index_name
        else:
            if self.index_name is None:
                print(
                    "Cannot search without an index_name! Please provide one.",
                    "Returning empty results.",
                )
                return None
        print(
            f"Loading searcher for index {self.index_name} for the first time...",
            "This may take a few seconds",
        )
        self.searcher = Searcher(
            checkpoint=self.checkpoint,
            config=None,
            collection=self.collection,
            index_root=self.config.root,
            index=self.index_name,
        )

        if not force_fast:
            if len(self.searcher.collection) < 10000:
                self.searcher.configure(ncells=4)
                self.searcher.configure(centroid_score_threshold=0.4)
                self.searcher.configure(ndocs=512)
            elif len(self.searcher.collection) < 100000:
                self.searcher.configure(ncells=2)
                self.searcher.configure(centroid_score_threshold=0.45)
                self.searcher.configure(ndocs=1024)
            # Otherwise, use defaults for k
        else:
            # Use fast settingss
            self.searcher.configure(ncells=1)
            self.searcher.configure(centroid_score_threshold=0.5)
            self.searcher.configure(ndocs=256)

        print("Searcher loaded!")

    def search(
        self,
        query: Union[str, list[str]],
        index_name: Optional["str"] = None,
        k: int = 10,
        force_fast: bool = False,
        zero_index_ranks: bool = False,
    ):
        if self.searcher is None or (
            index_name is not None and self.index_name != index_name
        ):
            self._load_searcher(index_name=index_name, force_fast=force_fast)

        if isinstance(query, str):
            results = [self._search(query, k)]
        else:
            results = self._batch_search(query, k)

        to_return = []

        for result in results:
            result_for_query = []
            for id_, rank, score in zip(*result):
                document_id = self.pid_docid_map[id_]
                result_dict = {
                    "content": self.collection[id_],
                    "score": score,
                    "rank": rank - 1 if zero_index_ranks else rank,
                    "document_id": document_id,
                }

                if self.docid_metadata_map is not None:
                    if document_id in self.docid_metadata_map:
                        doc_metadata = self.docid_metadata_map[document_id]
                        result_dict["document_metadata"] = doc_metadata

                result_for_query.append(result_dict)

            to_return.append(result_for_query)

        if len(to_return) == 1:
            return to_return[0]
        return to_return

    def _search(self, query: str, k: int):
        return self.searcher.search(query, k=k)

    def _batch_search(self, query: list[str], k: int):
        queries = {i: x for i, x in enumerate(query)}
        results = self.searcher.search_all(queries, k=k)
        results = [
            [list(zip(*value))[i] for i in range(3)]
            for value in results.todict().values()
        ]
        return results

    def train(self, data_dir, training_config: ColBERTConfig):
        training_config = ColBERTConfig.from_existing(self.config, training_config)
        training_config.nway = 2
        with Run().context(self.run_config):
            trainer = Trainer(
                triples=str(data_dir / "triples.train.colbert.jsonl"),
                queries=str(data_dir / "queries.train.colbert.tsv"),
                collection=str(data_dir / "corpus.train.colbert.tsv"),
                config=training_config,
            )

            trainer.train(checkpoint=self.checkpoint)

    def _colbert_score(self, Q, D_padded, D_mask):
        if ColBERTConfig().total_visible_gpus > 0:
            Q, D_padded, D_mask = Q.cuda(), D_padded.cuda(), D_mask.cuda()

        assert Q.dim() == 3, Q.size()
        assert D_padded.dim() == 3, D_padded.size()
        assert Q.size(0) in [1, D_padded.size(0)]

        scores = D_padded @ Q.to(dtype=D_padded.dtype).permute(0, 2, 1)
        scores = scores.max(1).values
        return scores.sum(-1)

    def _index_free_search(
        self,
        embedded_queries,
        documents: list[str],
        embedded_docs,
        doc_mask,
        k: int = 10,
        zero_index: bool = False,
    ):
        results = []

        for query in embedded_queries:
            results_for_query = []
            scores = self._colbert_score(query, embedded_docs, doc_mask)
            sorted_scores = sorted(enumerate(scores), key=lambda x: x[1], reverse=True)
            high_score_idxes = [index for index, _ in sorted_scores[:k]]
            for rank, doc_idx in enumerate(high_score_idxes):
                result = {
                    "content": documents[doc_idx],
                    "score": float(scores[doc_idx]),
                    "rank": rank - 1 if zero_index else rank,
                    "result_index": doc_idx,
                }
                results_for_query.append(result)
            results.append(results_for_query)

        if len(results) == 1:
            return results[0]

        return results

    def _set_inference_max_tokens(
        self, documents: list[str], max_tokens: Union[Literal["auto"], int] = "auto"
    ):
        if (
            not hasattr(self, "inference_ckpt_len_set")
            or self.inference_ckpt_len_set is False
        ):
            if max_tokens == "auto" or max_tokens > 512:
                max_tokens = 512
                percentile_90 = np.percentile(
                    [len(x.split(" ")) for x in documents], 90
                )
                max_tokens = min(
                    (math.ceil((percentile_90 * 1.35) / 32) * 32) * 1.1,
                    512,
                )
                max_tokens = max(256, max_tokens)
                if max_tokens > 288:
                    print(
                        f"Your documents are roughly {percentile_90} tokens long at the 90th percentile!",
                        "This is quite long and might slow down reranking!\n",
                        "Provide fewer documents, build smaller chunks or run on GPU",
                        "if it takes too long for your needs!",
                    )
            self.inference_ckpt.colbert_config.max_doclen = max_tokens
            self.inference_ckpt.doc_tokenizer.doc_maxlen = max_tokens
            self.inference_ckpt_len_set = True

    def _index_free_retrieve(
        self,
        query: Union[str, list[str]],
        documents: list[str],
        k: int,
        max_tokens: Union[Literal["auto"], int] = "auto",
        zero_index: bool = False,
        bsize: int = 32,
    ):
        self._set_inference_max_tokens(documents=documents, max_tokens=max_tokens)

        if k > len(documents):
            print("k value cannot be larger than the number of documents! aborting...")
            return None
        if len(documents) > 1000:
            print(
                "Please note ranking in-memory is not optimised for large document counts! ",
                "Consider building an index and using search instead!",
            )
        if len(set(documents)) != len(documents):
            print(
                "WARNING! Your documents have duplicate entries! ",
                "This will slow down calculation and may yield subpar results",
            )

        embedded_queries = self._encode_index_free_queries(query, bsize=bsize)
        embedded_docs, doc_mask = self._encode_index_free_documents(
            documents, bsize=bsize
        )

        return self._index_free_search(
            embedded_queries=embedded_queries,
            documents=documents,
            embedded_docs=embedded_docs,
            doc_mask=doc_mask,
            k=k,
            zero_index=zero_index,
        )

    def _encode_index_free_queries(
        self, queries: Union[str, list[str]], bsize: int = 32
    ):
        if isinstance(queries, str):
            queries = [queries]
        embedded_queries = [
            x.unsqueeze(0)
            for x in self.inference_ckpt.queryFromText(queries, bsize=bsize)
        ]
        return embedded_queries

    def _encode_index_free_documents(
        self, documents: list[str], bsize: int = 32, verbose: bool = True
    ):
        embedded_docs = self.inference_ckpt.docFromText(
            documents, bsize=bsize, showprogress=verbose
        )[0]
        doc_mask = torch.full(embedded_docs.shape[:2], -float("inf"))
        return embedded_docs, doc_mask

    def rank(
        self,
        query: str,
        documents: list[str],
        k: int = 10,
        zero_index_ranks: bool = False,
        bsize: int = 32,
    ):
        return self._index_free_retrieve(
            query, documents, k, zero_index=zero_index_ranks, bsize=bsize
        )

    def encode(
        self,
        documents: list[str],
        document_metadatas: Optional[list[dict]] = None,
        bsize: int = 32,
        max_tokens: Union[Literal["auto"], int] = "auto",
        verbose: bool = True,
    ):
        self._set_inference_max_tokens(documents=documents, max_tokens=max_tokens)
        encodings, doc_masks = self._encode_index_free_documents(
            documents, bsize=bsize, verbose=verbose
        )
        encodings = torch.cat(
            [
                encodings,
                torch.zeros(
                    (
                        encodings.shape[0],
                        self.inference_ckpt.doc_tokenizer.doc_maxlen
                        - encodings.shape[1],
                        encodings.shape[2],
                    )
                ),
            ],
            dim=1,
        )
        doc_masks = torch.cat(
            [
                doc_masks,
                torch.full(
                    (
                        doc_masks.shape[0],
                        self.inference_ckpt.colbert_config.max_doclen
                        - doc_masks.shape[1],
                    ),
                    -float("inf"),
                ),
            ],
            dim=1,
        )

        print("Shapes:")
        print(f"encodings: {encodings.shape}")
        print(f"doc_masks: {doc_masks.shape}")

        if hasattr(self, "in_memory_collection"):
            if self.in_memory_metadata is not None:
                if document_metadatas is None:
                    self.in_memory_metadatas.extend([None] * len(documents))
                else:
                    self.in_memory_metadata.extend(document_metadatas)
            elif document_metadatas is not None:
                self.in_memory_metadata = [None] * len(self.in_memory_collection)
                self.in_memory_metadata.extend(document_metadatas)

            self.in_memory_collection.extend(documents)

            # add 0 padding to encodings so they're self.inference_ckpt.doc_tokenizer.doc_maxlen length

            self.in_memory_embed_docs = torch.cat(
                [self.in_memory_embed_docs, encodings], dim=0
            )
            self.doc_masks = torch.cat([self.doc_masks, doc_masks], dim=0)

        else:
            self.in_memory_collection = documents
            self.in_memory_metadata = document_metadatas
            self.in_memory_embed_docs = encodings
            self.doc_masks = doc_masks

    def search_encoded_docs(
        self,
        queries: Union[str, list[str]],
        k: int = 10,
        bsize: int = 32,
    ):
        queries = self._encode_index_free_queries(queries, bsize=bsize)
        results = self._index_free_search(
            embedded_queries=queries,
            documents=self.in_memory_collection,
            embedded_docs=self.in_memory_embed_docs,
            doc_mask=self.doc_masks,
            k=k,
        )
        if self.in_memory_metadata is not None:
            for result in results:
                result["document_metadata"] = self.in_memory_metadata[
                    result["result_index"]
                ]
        return results

    def clear_encoded_docs(self, force: bool = False):
        if not force:
            print(
                "All in-memory encodings will be deleted in 10 seconds, interrupt now if you want to keep them!"
            )
            print("...")
            time.sleep(10)
        del self.in_memory_collection
        del self.in_memory_metadata
        del self.in_memory_embed_docs
        del self.doc_masks
        del self.inference_ckpt_len_set

    def __del__(self):
        # Clean up context
        self.run_context.__exit__(None, None, None)



from pathlib import Path
from typing import Literal, Optional, Union

from colbert.infra import ColBERTConfig

from ragatouille.data import TrainingDataProcessor
from ragatouille.models import ColBERT, LateInteractionModel
from ragatouille.negative_miners import HardNegativeMiner, SimpleMiner
from ragatouille.utils import seeded_shuffle


class RAGTrainer:
    """Main trainer to fine-tune/train ColBERT models with a few lines."""

    model: Union[LateInteractionModel, None] = None
    negative_miner: Union[HardNegativeMiner, None] = None
    collection: list[str] = []
    queries: Union[list[str], None] = None
    raw_data: Union[list[tuple], list[list], None] = None
    training_triplets: list[list[int]] = list()

    def __init__(
        self,
        model_name: str,
        pretrained_model_name: str,
        language_code: str = "en",
        n_usable_gpus: int = -1,
    ):
        """
        Initialise a RAGTrainer instance. This will load a base model: either an existing ColBERT model to fine-tune or a BERT/RoBERTa-like model to build a new ColBERT model from.

        Parameters:
            model_name: str - Name of the model to train. This will be used to name the checkpoints and the index.
            pretrained_model_name: str - Name of the pretrained model to use as a base. Can be a local path to a checkpoint or a huggingface model name.
            language_code: str - Language code of the model to train. This will be used to name the checkpoints and the index.
            n_usable_gpus: int - Number of GPUs to use. By default, value is -1, which means use all available GPUs or none if no GPU is available.

        Returns:
            self (RAGTrainer): The current instance of RAGTrainer, with the base model initialised.
        """

        self.model_name = model_name
        self.pretrained_model_name = pretrained_model_name
        self.language_code = language_code
        self.model = ColBERT(
            pretrained_model_name_or_path=pretrained_model_name,
            n_gpu=n_usable_gpus,
            training_mode=True,
        )

    def add_documents(self, documents: list[str]):
        self.collection += documents
        seeded_shuffle(self.collection)

    def export_training_data(self, path: Union[str, Path]):
        """
        Manually export the training data processed by prepare_training_data to a given path.

        Parameters:
            path: Union[str, Path] - Path to the directory where the data will be exported."""
        self.data_processor.export_training_data(path)

    def prepare_training_data(
        self,
        raw_data: Union[list[tuple], list[list]],
        all_documents: Optional[list[str]] = None,
        data_out_path: Union[str, Path] = "./data/",
        num_new_negatives: int = 10,
        hard_negative_minimum_rank: int = 10,
        mine_hard_negatives: bool = True,
        hard_negative_model_size: str = "small",
        pairs_with_labels: bool = False,
        positive_label: Union[int, str] = 1,
        negative_label: Union[int, str] = 0,
    ) -> str:
        """
        Fully pre-process input-data in various raw formats into ColBERT-ready files and triplets.
        Will accept a variety of formats, such as unannotated pairs, annotated pairs, triplets of strings and triplets of list of strings.
        Will process into a ColBERT-ready format and export to data_out_path.
        Will generate hard negatives if mine_hard_negatives is True.
        num_new_negatives decides how many negatives will be generated. if mine_hard_negatives is False and num_new_negatives is > 0, these negatives will be randomly sampled.

        Parameters:
            raw_data: Union[list[tuple], list[list]] - List of pairs, annotated pairs, or triplets of strings.
            all_documents: Optional[list[str]] - A corpus of documents to be used for sampling negatives.
            data_out_path: Union[str, Path] - Path to the directory where the data will be exported (can be a tmp directory).
            num_new_negatives: int - Number of new negatives to generate for each query.
            mine_hard_negatives: bool - Whether to use hard negatives mining or not.
            hard_negative_model_size: str - Size of the model to use for hard negatives mining.
            pairs_with_labels: bool - Whether the raw_data is a list of pairs with labels or not.
            positive_label: Union[int, str] - Label to use for positive pairs.
            negative_label: Union[int, str] - Label to use for negative pairs.

        Returns:
            data_out_path: Union[str, Path] - Path to the directory where the data has been exported.
        """
        if all_documents is not None:
            self.collection += all_documents

        self.data_dir = Path(data_out_path)
        if len(raw_data[0]) == 2:
            data_type = "pairs"
            if pairs_with_labels:
                data_type = "labeled_pairs"
        elif len(raw_data[0]) == 3:
            data_type = "triplets"
        else:
            raise ValueError("Raw data must be a list of pairs or triplets of strings.")

        if type(raw_data[0][1]) == str:
            self.collection += [x[1] for x in raw_data]
        else:
            for x in raw_data:
                for txt in x[1]:
                    self.collection.append(txt)
        if data_type == "triplets":
            if type(raw_data[0][2]) == str:
                self.collection += [x[2] for x in raw_data]
            else:
                for x in raw_data:
                    for txt in x[2]:
                        self.collection.append(txt)

        self.queries = set([x[0] for x in raw_data])
        self.collection = list(set(self.collection))
        seeded_shuffle(self.collection)

        if mine_hard_negatives:
            self.negative_miner = SimpleMiner(
                language_code=self.language_code,
                model_size=hard_negative_model_size,
            )
            self.negative_miner.build_index(self.collection)

        self.data_processor = TrainingDataProcessor(
            collection=self.collection,
            queries=self.queries,
            negative_miner=self.negative_miner if mine_hard_negatives else None,
        )

        self.data_processor.process_raw_data(
            data_type=data_type,
            raw_data=raw_data,
            export=True,
            data_dir=data_out_path,
            num_new_negatives=num_new_negatives,
            positive_label=positive_label,
            negative_label=negative_label,
            mine_hard_negatives=mine_hard_negatives,
            hard_negative_minimum_rank=hard_negative_minimum_rank,
        )

        self.training_triplets = self.data_processor.training_triplets

        return data_out_path

    def train(
        self,
        batch_size: int = 32,
        nbits: int = 2,
        maxsteps: int = 500_000,
        use_ib_negatives: bool = True,
        learning_rate: float = 5e-6,
        dim: int = 128,
        doc_maxlen: int = 256,
        use_relu: bool = False,
        warmup_steps: Union[int, Literal["auto"]] = "auto",
        accumsteps: int = 1,
    ) -> str:
        """
        Launch training or fine-tuning of a ColBERT model.

        Parameters:
            batch_size: int - Total batch size -- divice by n_usable_gpus for per-GPU batch size.
            nbits: int - number of bits used for vector compression by the traiened model. 2 is usually ideal.
            maxsteps: int - End training early after maxsteps steps.
            use_ib_negatives: bool - Whether to use in-batch negatives to calculate loss or not.
            learning_rate: float - ColBERT litterature usually has this performing best between 3e-6 - 2e-5 depending on data size
            dim: int - Size of individual vector representations.
            doc_maxlen: int - The maximum length after which passages will be truncated
            warmup_steps: Union[int, Literal["auto"]] - How many warmup steps to use for the learning rate.
                                                      Auto will default to 10% of total steps
            accumsteps: How many gradient accummulation steps to use to simulate higher batch sizes.

        Returns:
            model_path: str - Path to the trained model.
        """
        if not self.training_triplets:
            total_triplets = sum(
                1 for _ in open(str(self.data_dir / "triples.train.colbert.jsonl"), "r")
            )
        else:
            total_triplets = len(self.training_triplets)

        training_config = ColBERTConfig(
            bsize=batch_size,
            model_name=self.model_name,
            name=self.model_name,
            checkpoint=self.pretrained_model_name,
            use_ib_negatives=use_ib_negatives,
            maxsteps=maxsteps,
            nbits=nbits,
            lr=learning_rate,
            dim=dim,
            doc_maxlen=doc_maxlen,
            relu=use_relu,
            accumsteps=accumsteps,
            warmup=int(total_triplets // batch_size * 0.1)
            if warmup_steps == "auto"
            else warmup_steps,
            save_every=int(total_triplets // batch_size // 10),
        )

        return self.model.train(data_dir=self.data_dir, training_config=training_config)



from typing import Callable, Optional, Union
from uuid import uuid4

from ragatouille.data.preprocessors import llama_index_sentence_splitter


class CorpusProcessor:
    def __init__(
        self,
        document_splitter_fn: Optional[Callable] = llama_index_sentence_splitter,
        preprocessing_fn: Optional[Union[Callable, list[Callable]]] = None,
    ):
        self.document_splitter_fn = document_splitter_fn
        self.preprocessing_fn = preprocessing_fn

    def process_corpus(
        self,
        documents: list[str],
        document_ids: Optional[list[str]] = None,
        **splitter_kwargs,
    ) -> list[str]:
        # TODO CHECK KWARGS
        document_ids = (
            [str(uuid4()) for _ in range(len(documents))]
            if document_ids is None
            else document_ids
        )
        if self.document_splitter_fn is not None:
            documents = self.document_splitter_fn(
                documents, document_ids, **splitter_kwargs
            )
        if self.preprocessing_fn is not None:
            if isinstance(self.preprocessing_fn, list):
                for fn in self.preprocessing_fn:
                    documents = fn(documents, document_ids)
                return documents
            return self.preprocessing_fn(documents, document_ids)
        return documents



import os
import random
import shutil
from pathlib import Path
from typing import Union

import torch
import torch.nn as nn
from colbert.infra import ColBERTConfig
from colbert.modeling.colbert import ColBERT
from huggingface_hub import HfApi
from huggingface_hub.utils import HfHubHTTPError
from transformers import AutoModel, BertPreTrainedModel


def seeded_shuffle(collection: list, seed: int = 42):
    random.seed(seed)
    random.shuffle(collection)
    return collection


""" HUGGINGFACE """


def export_to_huggingface_hub(
    colbert_path: Union[str, Path],
    huggingface_repo_name: str,
    export_vespa_onnx: bool = False,
    use_tmp_dir: bool = False,
):
    # ensure model contains a valid ColBERT config before exporting
    colbert_config = ColBERTConfig.load_from_checkpoint(colbert_path)
    try:
        assert colbert_config is not None
    except Exception:
        print(f"Path {colbert_path} does not contain a valid ColBERT config!")

    export_path = colbert_path
    if use_tmp_dir:
        export_path = ".tmp/hugging_face_export"
        print("Using tmp dir to store export files...")
        colbert_model = ColBERT(
            colbert_path,
            colbert_config=colbert_config,
        )
        print(f"Model loaded... saving export files to disk at {export_path}")
        try:
            save_model = colbert_model.save
        except Exception:
            save_model = colbert_model.module.save
        save_model(export_path)

    if export_vespa_onnx:
        rust_tokenizer_available = True
        if use_tmp_dir:
            try:
                colbert_model.raw_tokenizer.save_pretrained(
                    export_path, legacy_format=False
                )
            except Exception:
                rust_tokenizer_available = False
        else:
            rust_tokenizer_available = os.path.exists(
                Path(colbert_path) / "tokenizer.json"
            )
        if not rust_tokenizer_available:
            print(
                "The tokenizer for your model does not seem to have a Fast Tokenizer implementation...\n",
                "This may cause problems when trying to use with Vespa!\n",
                "Proceeding anyway...",
            )

        export_to_vespa_onnx(colbert_path, out_path=export_path)
    try:
        api = HfApi()
        api.create_repo(repo_id=huggingface_repo_name, repo_type="model", exist_ok=True)
        api.upload_folder(
            folder_path=export_path,
            repo_id=huggingface_repo_name,
            repo_type="model",
        )
        print(f"Successfully uploaded model to {huggingface_repo_name}")
    except ValueError as e:
        print(
            f"Could not create repository on the huggingface hub.\n",
            f"Error: {e}\n",
            "Please make sure you are logged in (run huggingface-cli login)\n",
            "If the error persists, please open an issue on github. This is a beta feature!",
        )
    except HfHubHTTPError:
        print(
            "You don't seem to have the rights to create a repository with this name...\n",
            "Please make sure your repo name is in the format 'yourusername/your-repo-name'",
        )
    finally:
        if use_tmp_dir:
            shutil.rmtree(export_path)


""" VESPA """


class VespaColBERT(BertPreTrainedModel):
    def __init__(self, config, dim):
        super().__init__(config)
        self.bert = AutoModel.from_config(config)
        self.linear = nn.Linear(config.hidden_size, dim, bias=False)
        self.init_weights()

    def forward(self, input_ids, attention_mask):
        Q = self.bert(input_ids, attention_mask=attention_mask)[0]
        Q = self.linear(Q)
        return torch.nn.functional.normalize(Q, p=2, dim=2)


def export_to_vespa_onnx(
    colbert_path: Union[str, Path],
    out_path: Union[str, Path],
    out_file_name: str = "vespa_colbert.onnx",
):
    print(f"Exporting model {colbert_path} to {out_path}/{out_file_name}")
    out_path = Path(out_path)
    vespa_colbert = VespaColBERT.from_pretrained(colbert_path, dim=128)
    print("Model loaded, converting to ONNX...")
    input_names = ["input_ids", "attention_mask"]
    output_names = ["contextual"]
    input_ids = torch.ones(1, 32, dtype=torch.int64)
    attention_mask = torch.ones(1, 32, dtype=torch.int64)
    args = (input_ids, attention_mask)
    torch.onnx.export(
        vespa_colbert,
        args=args,
        f=str(out_path / out_file_name),
        input_names=input_names,
        output_names=output_names,
        dynamic_axes={
            "input_ids": {0: "batch", 1: "batch"},
            "attention_mask": {0: "batch", 1: "batch"},
            "contextual": {0: "batch", 1: "batch"},
        },
        opset_version=17,
    )
    print("Vespa ONNX export complete!")



from ragatouille.integrations._langchain import (
    RAGatouilleLangChainCompressor,
    RAGatouilleLangChainRetriever,
)

__all__ = ["RAGatouilleLangChainRetriever", "RAGatouilleLangChainCompressor"]



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union


class HardNegativeMiner(ABC):
    @abstractmethod
    def export_index(self, path: Union[str, Path]) -> bool:
        ...

    @abstractmethod
    def mine_hard_negatives(
        self,
        queries: list[str],
        collection: list[str],
        neg_k: int,
    ):
        ...

    @abstractmethod
    def _mine(
        self,
        queries: list[str],
        k: int,
    ):
        ...



import os
import random
from collections import defaultdict
from itertools import product
from pathlib import Path
from typing import Literal, Union

import srsly


class TrainingDataProcessor:
    def __init__(
        self,
        collection: list[str],
        queries: list[str],
        negative_miner=None,
    ):
        self.collection = collection
        self.queries = queries
        self.negative_miner = negative_miner
        self._make_data_map()
        self.training_triplets = []

    def process_raw_data(
        self,
        raw_data,
        data_type: Literal["pairs", "triplets", "labeled_pairs"],
        data_dir: Union[str, Path],
        export: bool = True,
        mine_hard_negatives: bool = True,
        num_new_negatives: int = 10,
        positive_label: int = 1,
        negative_label: int = 0,
        hard_negative_minimum_rank: int = 10,
    ):
        self.negative_miner.min_rank = hard_negative_minimum_rank
        if self.negative_miner is None and mine_hard_negatives:
            raise ValueError(
                "mine_hard_negatives is True but no negative miner was provided!"
            )
        if data_type == "pairs":
            self._process_raw_pairs(
                raw_data=raw_data,
                mine_hard_negatives=mine_hard_negatives,
                n_new_negatives=num_new_negatives,
            )
        elif data_type == "labeled_pairs":
            self._process_raw_labeled_pairs(
                raw_data=raw_data,
                mine_hard_negatives=mine_hard_negatives,
                n_new_negatives=num_new_negatives,
                positive_label=positive_label,
                negative_label=negative_label,
            )
        elif data_type == "triplets":
            self._process_raw_triplets(
                raw_data=raw_data,
                mine_hard_negatives=mine_hard_negatives,
                n_new_negatives=num_new_negatives,
            )

        if export:
            self.export_training_data(data_dir)

    def _make_individual_triplets(self, query, positives, negatives):
        """Create the training data in ColBERT(v1) format from raw lists of triplets"""
        triplets = []
        q = self.query_map[query]
        random.seed(42)
        if len(positives) > 1:
            all_pos_texts = [p for p in positives]
            max_triplets_per_query = 20
            negs_per_positive = max(1, max_triplets_per_query // len(all_pos_texts))
            initial_triplets_count = 0
            for pos in all_pos_texts:
                p = self.passage_map[pos]
                chosen_negs = random.sample(
                    negatives, min(len(negatives), negs_per_positive)
                )
                for neg in chosen_negs:
                    n = self.passage_map[neg]
                    initial_triplets_count += 1
                    triplets.append([q, p, n])

            extra_triplets_needed = max_triplets_per_query - initial_triplets_count
            if extra_triplets_needed > 0:
                all_combinations = list(product(all_pos_texts, negatives))
                random.seed(42)
                random.shuffle(all_combinations)
                for pos, neg in all_combinations:
                    p = self.passage_map[pos]
                    n = self.passage_map[neg]
                    if [q, p, n] not in triplets:
                        triplets.append([q, p, n])
                        extra_triplets_needed -= 1
                        if extra_triplets_needed <= 0:
                            break

        else:
            p = self.passage_map[positives[0]]
            for n in negatives:
                triplets.append([q, p, self.passage_map[n]])

        return triplets

    def _get_new_negatives(self, query, passages, mine_hard_negatives, n_new_negatives):
        """Generate new negatives for each query, using either:
        - The assigned hard negative miner if mine_hard_negatives is True
        - Randomly sampling from the full collection otherwise
        """
        if mine_hard_negatives:
            hard_negatives = self.negative_miner.mine_hard_negatives(query)
            candidates = [
                x
                for x in hard_negatives
                if x not in passages["positives"] and x not in passages["negatives"]
            ]
            new_negatives = random.sample(
                candidates,
                min(n_new_negatives, len(candidates)),
            )
        else:
            new_negatives = [
                x
                for x in random.sample(self.collection, n_new_negatives)
                if x not in passages["positives"] and x not in passages["negatives"]
            ]

        return new_negatives

    def _process_raw_pairs(self, raw_data, mine_hard_negatives, n_new_negatives):
        """Convert unlabeled pairs into training triplets.
        It's assumed unlabeled pairs are always in the format (query, relevant_passage)"""
        training_triplets = []
        raw_grouped_triplets = defaultdict(lambda: defaultdict(list))

        for query, positive in raw_data:
            if isinstance(positive, str):
                positive = [positive]
            raw_grouped_triplets[query]["positives"] += positive

        for query, passages in raw_grouped_triplets.items():
            if n_new_negatives > 0:
                passages["negatives"] += self._get_new_negatives(
                    query=query,
                    passages=passages,
                    mine_hard_negatives=mine_hard_negatives,
                    n_new_negatives=n_new_negatives,
                )
            training_triplets += self._make_individual_triplets(
                query=query,
                positives=list(set(passages["positives"])),
                negatives=list(set(passages["negatives"])),
            )
        self.training_triplets = training_triplets

    def _process_raw_labeled_pairs(
        self,
        raw_data,
        mine_hard_negatives,
        n_new_negatives,
        positive_label,
        negative_label,
    ):
        """
        Convert labeled pairs intro training triplets.
        Labeled pairs are in the format (query, passage, label)
        """
        training_triplets = []
        raw_grouped_triplets = defaultdict(lambda: defaultdict(list))

        for query, passage, label in raw_data:
            if isinstance(passage, str):
                passage = [passage]
            if label == positive_label:
                label = "positives"
            elif label == negative_label:
                label = "negatives"
            else:
                raise ValueError(
                    f"Label {label} must correspond to either positive_label or negative_label!"
                )

            raw_grouped_triplets[query][label] += passage

        for query, passages in raw_grouped_triplets.items():
            if n_new_negatives > 0:
                passages["negatives"] += self._get_new_negatives(
                    query=query,
                    passages=passages,
                    mine_hard_negatives=mine_hard_negatives,
                    n_new_negatives=n_new_negatives,
                )

            training_triplets += self._make_individual_triplets(
                query=query,
                positives=passages["positives"],
                negatives=passages["negatives"],
            )
        self.training_triplets = training_triplets

    def _process_raw_triplets(self, raw_data, mine_hard_negatives, n_new_negatives):
        """
        Convert raw triplets
        (query, positives : str | list[str], negatives: str | list[str])
        into training triplets.
        """
        training_triplets = []
        raw_grouped_triplets = defaultdict(lambda: defaultdict(list))
        for query, positive, negative in raw_data:
            if isinstance(positive, str):
                positive = [positive]
            if isinstance(negative, str):
                negative = [negative]

            raw_grouped_triplets[query]["positives"] += positive
            raw_grouped_triplets[query]["negatives"] += negative

        for query, passages in raw_grouped_triplets.items():
            if n_new_negatives > 0:
                passages["negatives"] += self._get_new_negatives(
                    query=query,
                    passages=passages,
                    mine_hard_negatives=mine_hard_negatives,
                    n_new_negatives=n_new_negatives,
                )
            training_triplets += self._make_individual_triplets(
                query=query,
                positives=passages["positives"],
                negatives=passages["negatives"],
            )
        self.training_triplets = training_triplets

    def _make_data_map(self):
        """
        Generate a query_text: query_id and passage_text: passage_id mapping
        To easily generate ColBERT-format training data.
        """
        self.query_map = {}
        self.passage_map = {}

        for i, query in enumerate(self.queries):
            self.query_map[query] = i
        for i, passage in enumerate(list(self.collection)):
            self.passage_map[passage] = i

    def export_training_data(self, path: Union[str, Path]):
        """
        Export training data for both training and versioning purposes.
        {path} should ideally be dvc versioned.
        """

        path = Path(path)

        # Create the directory if it does not exist
        os.makedirs(path, exist_ok=True)

        with open(path / "queries.train.colbert.tsv", "w") as f:
            for query, idx in self.query_map.items():
                query = query.replace("\t", " ").replace("\n", " ")
                f.write(f"{idx}\t{query}\n")
        with open(path / "corpus.train.colbert.tsv", "w") as f:
            for document, idx in self.passage_map.items():
                document = document.replace("\t", " ").replace("\n", " ")
                f.write(f"{idx}\t{document}\n")

        random.seed(42)
        random.shuffle(self.training_triplets)
        srsly.write_jsonl(path / "triples.train.colbert.jsonl", self.training_triplets)



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Union


class LateInteractionModel(ABC):
    @abstractmethod
    def __init__(
        self,
        pretrained_model_name_or_path: Union[str, Path],
        n_gpu,
    ):
        ...

    @abstractmethod
    def train():
        ...

    @abstractmethod
    def index(self, name: str, collection: list[str]):
        ...

    @abstractmethod
    def add_to_index(self):
        ...

    @abstractmethod
    def search(self, name: str, query: Union[str, list[str]]):
        ...

    @abstractmethod
    def _search(self, name: str, query: str):
        ...

    @abstractmethod
    def _batch_search(self, name: str, queries: list[str]):
        ...



from enum import Enum
from pathlib import Path
from typing import Literal, Optional, Union

import torch
from sentence_transformers import SentenceTransformer
from tqdm import tqdm
from voyager import Index, Space, StorageDataType

from .base import HardNegativeMiner


class DenseModels(Enum):
    en_small = "BAAI/bge-small-en-v1.5"
    zh_small = "thenlper/gte-small-zh"
    other_small = "intfloat/multilingual-e5-small"
    en_base = "BAAI/bge-base-en-v1.5"
    zh_base = "thenlper/gte-base-zh"
    other_base = "intfloat/multilingual-e5-base"
    en_large = "BAAI/bge-large-en-v1.5"
    zh_large = "thenlper/gte-large-zh"
    other_large = "intfloat/multilingual-e5-large"


class SimpleMiner(HardNegativeMiner):
    """The simplest approach to hard negatives mining.
    Select the most appropriate, small-sized embedding model for the target language.
    And retrieve random negatives in the top 10-100 results.
    Strong baseline for quick, low-engineering hard negative mining."""

    def __init__(
        self,
        language_code: str,
        model_size: Literal["small", "base", "large"] = "small",
    ) -> None:
        self.n_gpu = torch.cuda.device_count()
        self.target_language = language_code
        self.model_size = model_size
        if language_code not in ["en", "zh"]:
            language_code = "other"
        self.model_name = f"{language_code}_{model_size}"
        hub_model = DenseModels[self.model_name].value
        print(f"Loading Hard Negative SimpleMiner dense embedding model {hub_model}...")
        self.model = SentenceTransformer(hub_model)
        self.has_index = False
        self.min_rank = 10

    def build_index(
        self,
        collection,
        batch_size: int = 128,
        save_index: bool = False,
        save_path: Union[str, Path] = None,
        force_fp32: bool = True,
    ):
        print(f"Building hard negative index for {len(collection)} documents...")
        if len(collection) > 1000:
            pool = self.model.start_multi_process_pool()
            embeds = self.model.encode_multi_process(
                collection, pool, batch_size=batch_size
            )
            self.model.stop_multi_process_pool(pool)
        else:
            embeds = self.model.encode(collection, batch_size=batch_size)

        print("All documents embedded, now adding to index...")

        self.max_rank = min(110, int(len(collection) // 10))
        self.max_rank = min(self.max_rank, len(collection))

        storage_type = StorageDataType.Float32
        if len(collection) > 500000 and not force_fp32:
            storage_type = StorageDataType.E4M3

        self.voyager_index = Index(
            Space.Cosine,
            num_dimensions=self.model.get_sentence_embedding_dimension(),
            storage_data_type=storage_type,
        )

        self.corpus_map = {i: doc for i, doc in enumerate(collection)}
        id_to_vector = {}
        for i, emb in enumerate(embeds):
            id_to_vector[i] = emb
            self.corpus_map[i] = collection[i]
        del embeds

        self.voyager_index.add_items(
            vectors=[x for x in id_to_vector.values()],
            ids=[x for x in id_to_vector.keys()],
            num_threads=-1,
        )

        del id_to_vector

        if save_index:
            print(f"Saving index to {save_path}...")
            self.export_index(save_path)
        else:
            print("save_index set to False, skipping saving hard negative index")
        print("Hard negative index generated")
        self.has_index = True

    def query_index(self, query, top_k=110):
        results = self.voyager_index.query(
            query, k=min(top_k, self.voyager_index.__len__())
        )
        return results

    def mine_hard_negatives(
        self,
        queries: Union[list[str], str],
        collection: Optional[list[str]] = None,
        save_index: bool = False,
        save_path: Union[str, Path] = None,
        force_fp32: bool = True,
    ):
        if self.has_index is False and collection is not None:
            self.build_index(
                collection,
                save_index=save_index,
                save_path=save_path,
                force_fp32=force_fp32,
            )
        if isinstance(queries, str):
            return self._mine(queries)
        return self._batch_mine(queries)

    def _mine(
        self,
        query: str,
    ):
        q_emb = self.model.encode(query)
        query_results = self.query_index(q_emb, top_k=self.max_rank)
        if len(query_results) > self.min_rank:
            query_results = query_results[self.min_rank : self.max_rank]
        query_results = [self.corpus_map[x] for x in query_results[0]]
        return query_results

    def _batch_mine(
        self,
        queries: list[str],
    ):
        """Separate function to parallelise later on"""
        print(f"Retrieving hard negatives for {len(queries)} queries...")
        results = []
        print("Embedding queries...")
        query_embeddings = self.model.encode(queries, show_progress_bar=True)
        print("Retrieving hard negatives...")
        for q_emb in tqdm(query_embeddings):
            query_results = self.query_index(q_emb, top_k=self.max_rank)
            query_results = query_results[self.min_rank : self.max_rank]
            query_results = [self.corpus_map[x.id] for x in query_results]
            results.append(query_results)
        print(f"""Done generating hard negatives.""")
        return results

    def export_index(self, path: Union[str, Path]) -> bool:
        self.voyager_index.save(path)
        return True



from typing import Any, List, Optional, Sequence

from langchain.retrievers.document_compressors.base import BaseDocumentCompressor
from langchain_core.callbacks.manager import CallbackManagerForRetrieverRun, Callbacks
from langchain_core.documents import Document
from langchain_core.retrievers import BaseRetriever


class RAGatouilleLangChainRetriever(BaseRetriever):
    model: Any
    kwargs: dict = {}

    def _get_relevant_documents(
        self,
        query: str,
        *,
        run_manager: CallbackManagerForRetrieverRun,  # noqa
    ) -> List[Document]:
        """Get documents relevant to a query."""
        docs = self.model.search(query, **self.kwargs)
        return [Document(page_content=doc["content"]) for doc in docs]


class RAGatouilleLangChainCompressor(BaseDocumentCompressor):
    model: Any
    kwargs: dict = {}
    k: int = 5

    class Config:
        """Configuration for this pydantic object."""

        arbitrary_types_allowed = True

    def compress_documents(
        self,
        documents: Sequence[Document],
        query: str,
        callbacks: Optional[Callbacks] = None,  # noqa
        **kwargs,
    ) -> Any:
        """Rerank a list of documents relevant to a query."""
        doc_list = list(documents)
        _docs = [d.page_content for d in doc_list]
        results = self.model.rerank(
            query=query,
            documents=_docs,
            k=kwargs.get("k", self.k),
            **self.kwargs,
        )
        final_results = []
        for r in results:
            doc = doc_list[r["result_index"]]
            doc.metadata["relevance_score"] = r["score"]
            final_results.append(doc)
        return final_results



from pathlib import Path
from typing import Any, Callable, List, Literal, Optional, TypeVar, Union
from uuid import uuid4

from langchain.retrievers.document_compressors.base import BaseDocumentCompressor
from langchain_core.retrievers import BaseRetriever

from ragatouille.data.corpus_processor import CorpusProcessor
from ragatouille.data.preprocessors import llama_index_sentence_splitter
from ragatouille.integrations import (
    RAGatouilleLangChainCompressor,
    RAGatouilleLangChainRetriever,
)
from ragatouille.models import ColBERT, LateInteractionModel


class RAGPretrainedModel:
    """
    Wrapper class for a pretrained RAG late-interaction model, and all the associated utilities.
    Allows you to load a pretrained model from disk or from the hub, build or query an index.

    ## Usage

    Load a pre-trained checkpoint:

    ```python
    from ragatouille import RAGPretrainedModel

    RAG = RAGPretrainedModel.from_pretrained("colbert-ir/colbertv2.0")
    ```

    Load checkpoint from an existing index:

    ```python
    from ragatouille import RAGPretrainedModel

    RAG = RAGPretrainedModel.from_index("path/to/my/index")
    ```

    Both methods will load a fully initialised instance of ColBERT, which you can use to build and query indexes.

    ```python
    RAG.search("How many people live in France?")
    ```
    """

    model_name: Union[str, None] = None
    model: Union[LateInteractionModel, None] = None
    corpus_processor: Optional[CorpusProcessor] = None

    @classmethod
    def from_pretrained(
        cls,
        pretrained_model_name_or_path: Union[str, Path],
        n_gpu: int = -1,
        verbose: int = 1,
        index_root: Optional[str] = None,
    ):
        """Load a ColBERT model from a pre-trained checkpoint.

        Parameters:
            pretrained_model_name_or_path (str): Local path or huggingface model name.
            n_gpu (int): Number of GPUs to use. By default, value is -1, which means use all available GPUs or none if no GPU is available.
            verbose (int): The level of ColBERT verbosity requested. By default, 1, which will filter out most internal logs.
            index_root (Optional[str]): The root directory where indexes will be stored. If None, will use the default directory, '.ragatouille/'.

        Returns:
            cls (RAGPretrainedModel): The current instance of RAGPretrainedModel, with the model initialised.
        """
        instance = cls()
        instance.model = ColBERT(
            pretrained_model_name_or_path, n_gpu, index_root=index_root, verbose=verbose
        )
        return instance

    @classmethod
    def from_index(
        cls, index_path: Union[str, Path], n_gpu: int = -1, verbose: int = 1
    ):
        """Load an Index and the associated ColBERT encoder from an existing document index.

        Parameters:
            index_path (Union[str, path]): Path to the index.
            n_gpu (int): Number of GPUs to use. By default, value is -1, which means use all available GPUs or none if no GPU is available.
            verbose (int): The level of ColBERT verbosity requested. By default, 1, which will filter out most internal logs.

        Returns:
            cls (RAGPretrainedModel): The current instance of RAGPretrainedModel, with the model and index initialised.
        """
        instance = cls()
        index_path = Path(index_path)
        instance.model = ColBERT(
            index_path, n_gpu, verbose=verbose, load_from_index=True
        )

        return instance

    def _process_metadata(
        self,
        document_ids: Optional[Union[TypeVar("T"), List[TypeVar("T")]]],
        document_metadatas: Optional[list[dict[Any, Any]]],
        collection_len: int,
    ) -> tuple[list[str], Optional[dict[Any, Any]]]:
        if document_ids is None:
            document_ids = [str(uuid4()) for i in range(collection_len)]
        else:
            if len(document_ids) != collection_len:
                raise ValueError("document_ids must be the same length as collection")
            if len(document_ids) != len(set(document_ids)):
                raise ValueError("document_ids must be unique")
            if any(not id.strip() for id in document_ids):
                raise ValueError("document_ids must not contain empty strings")
            if not all(isinstance(id, type(document_ids[0])) for id in document_ids):
                raise ValueError("All document_ids must be of the same type")

        if document_metadatas is not None:
            if len(document_metadatas) != collection_len:
                raise ValueError(
                    "document_metadatas must be the same length as collection"
                )
            docid_metadata_map = {
                x: y for x, y in zip(document_ids, document_metadatas)
            }
        else:
            docid_metadata_map = None

        return document_ids, docid_metadata_map

    def index(
        self,
        collection: list[str],
        document_ids: Union[TypeVar("T"), List[TypeVar("T")]] = None,
        document_metadatas: Optional[list[dict]] = None,
        index_name: str = None,
        overwrite_index: Union[bool, str] = True,
        max_document_length: int = 256,
        split_documents: bool = True,
        document_splitter_fn: Optional[Callable] = llama_index_sentence_splitter,
        preprocessing_fn: Optional[Union[Callable, list[Callable]]] = None,
    ):
        """Build an index from a list of documents.

        Parameters:
            collection (list[str]): The collection of documents to index.
            document_ids (Optional[list[str]]): An optional list of document ids. Ids will be generated at index time if not supplied.
            index_name (str): The name of the index that will be built.
            overwrite_index (Union[bool, str]): Whether to overwrite an existing index with the same name.
            max_document_length (int): The maximum length of a document. Documents longer than this will be split into chunks.
            split_documents (bool): Whether to split documents into chunks.
            document_splitter_fn (Optional[Callable]): A function to split documents into chunks. If None and by default, will use the llama_index_sentence_splitter.
            preprocessing_fn (Optional[Union[Callable, list[Callable]]]): A function or list of functions to preprocess documents. If None and by default, will not preprocess documents.

        Returns:
            index (str): The path to the index that was built.
        """

        document_ids, docid_metadata_map = self._process_metadata(
            document_ids=document_ids,
            document_metadatas=document_metadatas,
            collection_len=len(collection),
        )

        if split_documents or preprocessing_fn is not None:
            self.corpus_processor = CorpusProcessor(
                document_splitter_fn=document_splitter_fn if split_documents else None,
                preprocessing_fn=preprocessing_fn,
            )
            collection_with_ids = self.corpus_processor.process_corpus(
                collection,
                document_ids,
                chunk_size=max_document_length,
            )
        else:
            collection_with_ids = [
                {"document_id": x, "content": y}
                for x, y in zip(document_ids, collection)
            ]

        pid_docid_map = {
            index: item["document_id"] for index, item in enumerate(collection_with_ids)
        }
        collection = [x["content"] for x in collection_with_ids]
        return self.model.index(
            collection,
            pid_docid_map=pid_docid_map,
            docid_metadata_map=docid_metadata_map,
            index_name=index_name,
            max_document_length=max_document_length,
            overwrite=overwrite_index,
        )

    def add_to_index(
        self,
        new_collection: list[str],
        new_document_ids: Optional[Union[TypeVar("T"), List[TypeVar("T")]]] = None,
        new_document_metadatas: Optional[list[dict]] = None,
        index_name: Optional[str] = None,
        split_documents: bool = True,
        document_splitter_fn: Optional[Callable] = llama_index_sentence_splitter,
        preprocessing_fn: Optional[Union[Callable, list[Callable]]] = None,
    ):
        """Add documents to an existing index.

        Parameters:
            new_collection (list[str]): The documents to add to the index.
            new_document_metadatas (Optional[list[dict]]): An optional list of metadata dicts
            index_name (Optional[str]): The name of the index to add documents to. If None and by default, will add documents to the already initialised one.
        """
        new_document_ids, new_docid_metadata_map = self._process_metadata(
            document_ids=new_document_ids,
            document_metadatas=new_document_metadatas,
            collection_len=len(new_collection),
        )

        if split_documents or preprocessing_fn is not None:
            self.corpus_processor = CorpusProcessor(
                document_splitter_fn=document_splitter_fn if split_documents else None,
                preprocessing_fn=preprocessing_fn,
            )
            new_collection_with_ids = self.corpus_processor.process_corpus(
                new_collection,
                new_document_ids,
                chunk_size=self.model.config.doc_maxlen,
            )
        else:
            new_collection_with_ids = [
                {"document_id": x, "content": y}
                for x, y in zip(new_document_ids, new_collection)
            ]

        new_collection = [x["content"] for x in new_collection_with_ids]

        new_pid_docid_map = {
            index: item["document_id"]
            for index, item in enumerate(new_collection_with_ids)
        }

        self.model.add_to_index(
            new_collection,
            new_pid_docid_map,
            new_docid_metadata_map=new_docid_metadata_map,
            index_name=index_name,
        )

    def delete_from_index(
        self,
        document_ids: Union[TypeVar("T"), List[TypeVar("T")]],
        index_name: Optional[str] = None,
    ):
        """Delete documents from an index by their IDs.

        Parameters:
            document_ids (Union[TypeVar("T"), List[TypeVar("T")]]): The IDs of the documents to delete.
            index_name (Optional[str]): The name of the index to delete documents from. If None and by default, will delete documents from the already initialised one.
        """
        self.model.delete_from_index(
            document_ids,
            index_name=index_name,
        )

    def search(
        self,
        query: Union[str, list[str]],
        index_name: Optional["str"] = None,
        k: int = 10,
        force_fast: bool = False,
        zero_index_ranks: bool = False,
        **kwargs,
    ):
        """Query an index.

        Parameters:
            query (Union[str, list[str]]): The query or list of queries to search for.
            index_name (Optional[str]): Provide the name of an index to query. If None and by default, will query an already initialised one.
            k (int): The number of results to return for each query.
            force_fast (bool): Whether to force the use of a faster but less accurate search method.
            zero_index_ranks (bool): Whether to zero the index ranks of the results. By default, result rank 1 is the highest ranked result

        Returns:
            results (Union[list[dict], list[list[dict]]]): A list of dict containing individual results for each query. If a list of queries is provided, returns a list of lists of dicts. Each result is a dict with keys `content`, `score`, `rank`, and 'document_id'. If metadata was indexed for the document, it will be returned under the "document_metadata" key.

        Individual results are always in the format:
        ```python3
        {"content": "text of the relevant passage", "score": 0.123456, "rank": 1, "document_id": "x"}
        ```
        or
        ```python3
        {"content": "text of the relevant passage", "score": 0.123456, "rank": 1, "document_id": "x", "document_metadata": {"metadata_key": "metadata_value", ...}}
        ```

        """
        return self.model.search(
            query=query,
            index_name=index_name,
            k=k,
            force_fast=force_fast,
            zero_index_ranks=zero_index_ranks,
            **kwargs,
        )

    def rerank(
        self,
        query: Union[str, list[str]],
        documents: list[str],
        k: int = 10,
        zero_index_ranks: bool = False,
        bsize: int = 64,
    ):
        """Encode documents and rerank them in-memory. Performance degrades rapidly with more documents.

        Parameters:
            query (Union[str, list[str]]): The query or list of queries to search for.
            documents (list[str]): The documents to rerank.
            k (int): The number of results to return for each query.
            zero_index_ranks (bool): Whether to zero the index ranks of the results. By default, result rank 1 is the highest ranked result
            bsize (int): The batch size to use for re-ranking.

        Returns:
            results (Union[list[dict], list[list[dict]]]): A list of dict containing individual results for each query. If a list of queries is provided, returns a list of lists of dicts. Each result is a dict with keys `content`, `score` and `rank`.

        Individual results are always in the format:
        ```python3
        {"content": "text of the relevant passage", "score": 0.123456, "rank": 1}
        ```
        """

        return self.model.rank(
            query=query,
            documents=documents,
            k=k,
            zero_index_ranks=zero_index_ranks,
            bsize=bsize,
        )

    def encode(
        self,
        documents: list[str],
        bsize: int = 32,
        document_metadatas: Optional[list[dict]] = None,
        verbose: bool = True,
        max_document_length: Union[Literal["auto"], int] = "auto",
    ):
        """Encode documents in memory to be searched through with no Index. Performance degrades rapidly with more documents.

        Parameters:
            documents (list[str]): The documents to encode.
            bsize (int): The batch size to use for encoding.
            document_metadatas (Optional[list[dict]]): An optional list of metadata dicts. Each entry must correspond to a document.
        """
        if verbose:
            print(f"Encoding {len(documents)} documents...")
        self.model.encode(
            documents=documents,
            bsize=bsize,
            document_metadatas=document_metadatas,
            verbose=verbose,
            max_tokens=max_document_length,
        )
        if verbose:
            print("Documents encoded!")

    def search_encoded_docs(
        self,
        query: Union[str, list[str]],
        k: int = 10,
        bsize: int = 32,
    ) -> list[dict[str, Any]]:
        """Search through documents encoded in-memory.

        Parameters:
            query (Union[str, list[str]]): The query or list of queries to search for.
            k (int): The number of results to return for each query.
            batch_size (int): The batch size to use for searching.

        Returns:
            results (list[dict[str, Any]]): A list of dict containing individual results for each query. If a list of queries is provided, returns a list of lists of dicts.
        """
        return self.model.search_encoded_docs(
            queries=query,
            k=k,
            bsize=bsize,
        )

    def clear_encoded_docs(self, force: bool = False):
        """Clear documents encoded in-memory.

        Parameters:
            force (bool): Whether to force the clearing of encoded documents without enforcing a 10s wait time.
        """
        self.model.clear_encoded_docs(force=force)

    def as_langchain_retriever(self, **kwargs: Any) -> BaseRetriever:
        return RAGatouilleLangChainRetriever(model=self, kwargs=kwargs)

    def as_langchain_document_compressor(
        self, k: int = 5, **kwargs: Any
    ) -> BaseDocumentCompressor:
        return RAGatouilleLangChainCompressor(model=self, k=k, kwargs=kwargs)



import random

import requests


def seeded_shuffle(collection: list, seed: int = 42):
    random.seed(seed)
    random.shuffle(collection)
    return collection


def get_wikipedia_page(title: str):
    """
    Retrieve the full text content of a Wikipedia page.

    :param title: str - Title of the Wikipedia page.
    :return: str - Full text content of the page as raw string.
    """
    # Wikipedia API endpoint
    URL = "https://en.wikipedia.org/w/api.php"

    # Parameters for the API request
    params = {
        "action": "query",
        "format": "json",
        "titles": title,
        "prop": "extracts",
        "explaintext": True,
    }

    # Custom User-Agent header to comply with Wikipedia's best practices
    headers = {"User-Agent": "RAGatouille_tutorial/0.0.1 (ben@clavie.eu)"}

    response = requests.get(URL, params=params, headers=headers)
    data = response.json()

    # Extracting page content
    page = next(iter(data["query"]["pages"].values()))
    return page["extract"] if "extract" in page else None



from llama_index import Document
from llama_index.text_splitter import SentenceSplitter


def llama_index_sentence_splitter(
    documents: list[str], document_ids: list[str], chunk_size=256
):
    chunk_overlap = min(chunk_size / 4, min(chunk_size / 2, 64))
    chunks = []
    node_parser = SentenceSplitter(chunk_size=chunk_size, chunk_overlap=chunk_overlap)
    docs = [[Document(text=doc)] for doc in documents]
    for doc_id, doc in zip(document_ids, docs):
        chunks += [
            {"document_id": doc_id, "content": node.text} for node in node_parser(doc)
        ]
    return chunks



from .corpus_processor import CorpusProcessor
from .preprocessors import llama_index_sentence_splitter
from .training_data_processor import TrainingDataProcessor

__all__ = [
    "TrainingDataProcessor",
    "CorpusProcessor",
    "llama_index_sentence_splitter",
]
